/* oe.h -- Header file for optimal estimation library

   Copyright (C) 2005 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef _OE_H
#define _OE_H 1

#include <stdlib.h>
#include <stdio.h>

#include "readconfig.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Error codes */
#define OE_SUCCESS 0
#define OE_FAILURE 1
#define OE_MEMORY_ERROR 2
#define OE_DIMENSION_ERROR 3
#define OE_MATRIX_ERROR 4
#define OE_RESIZE_PERMISSION_ERROR 5

/* Other codes */
#define OE_NO 0
#define OE_YES 1

/* Number of separate "a priori"-like constraints */
#define OE_MAX_PRIOR_CONSTRAINTS 4

/* Global variables containing backtrace information for error */
#define OE_MAX_ERROR_NESTING 10
extern char *oe_error_message;
extern char *oe_error_file_list[OE_MAX_ERROR_NESTING];
extern int oe_error_line_list[OE_MAX_ERROR_NESTING];
extern int oe_error_nesting;
extern int oe_error_code;

extern int oe_verbose;

#define OE_DEBUG fprintf(stderr, "Reached %s, line %d\n", __FILE__, __LINE__)

/* If an error occurs in a function in this library then the function
   will call OE_THROW(), passing the error code and error message to
   the caller */
#define OE_THROW(code, message) \
  do { \
    oe_error_message = message; \
    oe_error_file_list[0] = __FILE__; \
    oe_error_line_list[0] = __LINE__; \
    oe_error_nesting = 1; \
    return (code); \
  } while (0)

/* A wrapper around calls to functions in this library that checks the
   output and if an error occurred, it appends file and line data to
   the backtrace variables and returns to its own calling function */
#define OE_CHECK(status) \
  if ((oe_error_code = (status))) { \
    if (oe_error_nesting < OE_MAX_ERROR_NESTING) { \
      oe_error_file_list[oe_error_nesting] = __FILE__; \
      oe_error_line_list[oe_error_nesting] = __LINE__; \
    } \
    oe_error_nesting++; \
    return oe_error_code; \
  } \
  else { \
    oe_error_nesting = 0; \
    oe_error_message = NULL; \
  }

/* A wrapper around calls to functions that should be used in the
   top-level file; if an error occurs then it will print the backtrace
   information to standard error and exit the program */
#define OE_CATCH(status) \
  if ((oe_error_code = (status))) { \
    int i; \
    if (oe_error_message && oe_error_nesting) { \
      fprintf(stderr, "\nError: %s\n   at line %d of %s\n", \
	      oe_error_message, \
	      oe_error_line_list[0], oe_error_file_list[0]); \
      for (i = 0; i < oe_error_nesting && i < OE_MAX_ERROR_NESTING; i++) { \
        fprintf(stderr, "   called from line %d of %s\n", \
          oe_error_line_list[i], oe_error_file_list[i]); \
      } \
      if (oe_error_nesting > OE_MAX_ERROR_NESTING) { \
        fprintf(stderr, "   ...\n"); \
      } \
    } \
    else { \
      fprintf(stderr, "\nAn error occurred (code %d)\n", oe_error_code); \
    } \
    fprintf(stderr, "   called from line %d of %s\n" \
        "Exiting with error code %d\n", __LINE__, __FILE__, oe_error_code); \
    exit(oe_error_code); \
  } \
  else { \
    oe_error_nesting = 0; \
    oe_error_message = NULL; \
  }

/* A wrapper around calls to functions that should be used in the
   top-level file; if an error occurs then it will print the backtrace
   information to standard error and exit the program */
#define OE_WARN(status) \
  if ((oe_error_code = (status))) { \
    int i; \
    if (oe_error_message && oe_error_nesting) { \
      fprintf(stderr, "\nWarning: %s\n   at line %d of %s\n", \
	      oe_error_message, \
	      oe_error_line_list[0], oe_error_file_list[0]); \
      for (i = 0; i < oe_error_nesting && i < OE_MAX_ERROR_NESTING; i++) { \
        fprintf(stderr, "   called from line %d of %s\n", \
          oe_error_line_list[i], oe_error_file_list[i]); \
      } \
      if (oe_error_nesting > OE_MAX_ERROR_NESTING) { \
        fprintf(stderr, "   ...\n"); \
      } \
    } \
    else { \
      fprintf(stderr, "\nWarning: an error occurred (code %d)\n", oe_error_code); \
    } \
    fprintf(stderr, "   called from line %d of %s\n", __LINE__, __FILE__); \
  } \
  else { \
    oe_error_nesting = 0; \
    oe_error_message = NULL; \
  }

/* Use double precision */
typedef double OEreal;

/* Vector and matrix data structures for reals and integers */
typedef struct {
  OEreal *data;
  int n;
  int _n_allocated;
  OEreal missing_value;
} OEvector;

typedef struct {
  int *data;
  int n;
  int _n_allocated;
  int missing_value;
} OEintvector;

typedef struct {
  OEreal **data;
  int n, m;
  int _mn_allocated;
  int _m_allocated;
  OEreal missing_value;
} OEmatrix;

typedef struct {
  int **data;
  int n, m;
  int _mn_allocated;
  int _m_allocated;
  int missing_value;
} OEintmatrix;

/* An initialiser for new matrices and vectors */
#define OE_NEW_VECTOR {0, 0, 0, -999}
#define OE_NEW_MATRIX {0, 0, 0, 0, 0, -999}

/* An initialiser for a vector that is to point to a single row of a
   matrix */
#define OE_EXTRACT_ROW(M, irow) {(M)->data[irow], (M)->n, 0}

/* Debugging functions in oe_debug.c */
int oe_find_nans_in_vector(OEvector *v, char *name);
int oe_find_nans_in_matrix(OEmatrix *M, char *name);

/* Low-level memory allocation functions in oe_memory.c */
extern void (*oe_free) (void *);
extern void* (*oe_malloc) (size_t size);
extern void* (*oe_realloc) (void *ptr, size_t nbytes);
void oe_set_memory_functions(void* (*malloc_function) (size_t),
			     void* (*realloc_function) (void *, size_t),
			     void (*free_function) (void *));
char *oe_strdup(const char *s);

/* Memory allocation functions in oe_alloc.c */
int oe_resize_permitted(int new_permit);
int oe_zero_on_reset(int new_setting);
int oe_reset_vector(OEvector *v, int n);
int oe_reset_int_vector(OEintvector *v, int n);
int oe_reset_matrix(OEmatrix *M, int m, int n);
int oe_reset_int_matrix(OEintmatrix *M, int m, int n);

int oe_subset_matrix(OEmatrix *Mdest, OEmatrix *Msrc, OEintvector *irows,
		     int start_col, int end_col);
int oe_block_subset_matrix(OEmatrix *Mdest, OEmatrix *Msrc, int start_row,
			   int end_row, int start_col, int end_col);

void oe_free_vector(OEvector *v);
void oe_free_int_vector(OEintvector *v);
void oe_free_matrix(OEmatrix *M);
void oe_free_int_matrix(OEintmatrix *M);

/* Basic functions in oe_basic.c */
int oe_zero_vector(OEvector *v);
int oe_zero_int_vector(OEintvector *v);
int oe_zero_matrix(OEmatrix *M);
int oe_zero_int_matrix(OEintmatrix *M);

int oe_constant_vector(OEvector *v, OEreal value);
int oe_constant_matrix(OEmatrix *M, OEreal value);

int oe_copy_vector(OEvector *v_src, OEvector *v_dest);
int oe_copy_vector_part(OEvector *v_src, int src_start, int num,
			OEreal *dest);
int oe_copy_matrix(OEmatrix *M_src, OEmatrix *M_dest);
int oe_copy_matrix_part(OEmatrix *M_src,
			int src_m_start, int src_n_start,
			int num_m, int num_n,
			OEmatrix *M_dest,
			int dest_m_start, int dest_n_start);

int oe_reverse_vector(OEvector *v);

int oe_link_vector(OEvector *v, int n, OEreal *data);
int oe_link_matrix(OEmatrix *M, int m, int n, OEreal **data);

int oe_scale_vector(OEvector *v, OEreal scaling);
int oe_scale_matrix(OEmatrix *M, OEreal scaling);

/* Matrix multiplication functions in oe_multiply.c */
int oe_matrix_x_vector(OEmatrix *M, OEvector *v, OEvector *v_out);
int oe_matrixT_x_vector(OEmatrix *M, OEvector *v, OEvector *v_out);
int oe_diagonal_x_matrix(OEvector *v, OEmatrix *M, OEmatrix *M_out);
int oe_add_matrix(OEmatrix *src, OEmatrix *target);
int oe_subtract_matrix(OEmatrix *src, OEmatrix *target);
int oe_add_band_diagonal(OEmatrix *banddiag, int off_diagonals,
			 OEmatrix *target);
int oe_band_diagonal_x_vector(OEmatrix *M, OEvector *v, OEvector *v_out);
int oe_diagonal_x_band_diagonal(OEvector *v, OEmatrix *M, OEmatrix *M_out);
int oe_matrix_x_matrix(OEmatrix *M1, OEmatrix *M2, OEmatrix *M_out);
int oe_band_diagonal_x_matrix(OEmatrix *M1, OEmatrix *M2, OEmatrix *M_out);
int oe_matrix_x_band_diagonal(OEmatrix *M1, OEmatrix *M2, OEmatrix *M_out);

/* Functions for generating idealised error covariance matrices in
   oe_errcov.c */
int oe_exponential_errcov(int n, OEreal diagval, OEreal indexdecorr,
			  OEmatrix *M_out);
int oe_gaussian_errcov(int n, OEreal diagval, OEreal indexdecorr,
		       OEmatrix *M_out);

/* Functions for the preparation of matrices in oe_prepare.c */
int oe_prepare(OEvector *x,       /* State vector */
	       OEvector *y,       /* Measurements */
	       OEvector *y_model, /* Forward model simulated measurements */
	       OEmatrix *K,       /* Jacobian */
	       OEvector *inv_s,   /* Measurement error variances */
	       OEmatrix *A_out,   /* Hessian */
	       OEvector *rhs_out  /* Right-hand size of matrix equation */
	       );
int oe_register_a_priori(OEvector *xa, OEmatrix *Sa);
int oe_register_a_priori_invs(OEvector *xa, OEmatrix *invSa);
int oe_register_diagonal_a_priori(OEvector *xa, OEvector *sa);
int oe_register_diagonal_a_priori_invs(OEvector *xa, OEvector *inv_sa);
int oe_register_a_priori_i(int i, OEvector *xa, OEmatrix *Sa);
int oe_register_a_priori_invs_i(int i, OEvector *xa, OEmatrix *invSa);
int oe_register_diagonal_a_priori_i(int i, OEvector *xa, OEvector *sa);
int oe_register_diagonal_a_priori_invs_i(int i, OEvector *xa,
					 OEvector *inv_sa);
int oe_print_a_priori_status(FILE *file);
int oe_reset_smoothness_constraint(int n);
int oe_get_smoothness_constraint(OEmatrix **H);
int oe_register_smoothness_constraint(OEreal factor, int n1, int n2);
int oe_register_pinned_smoothness_constraint(OEreal factor, int n1, int n2);
int oe_register_flatness_constraint(OEreal factor, int n1, int n2);
int oe_chi_squared(OEvector *y, OEvector *y_model, OEvector *inv_s,
		   OEreal *chi_squared_out);
int oe_subtract_a_priori(OEmatrix *A, int jstart, int jend);

/* Functions for solving linear systems of equations in oe_solve.c */
int oe_solve(OEmatrix *A,  /* Hessian */
	     OEvector *rhs /* Right-hand side of matrix equation */
	     /* and update to state vector */
	     );
int oe_solve_symmetric(OEmatrix *A, OEvector *rhs);

int oe_invert(OEmatrix *A, OEmatrix *invA);
int oe_invert_symmetric(OEmatrix *A, OEmatrix *invA);

/* LU decomposition in oe_lu.c */
int oe_lu_decomposition(OEmatrix *A, int *index, int *interchanges_even);
int oe_lu_substitution(OEmatrix *A, int *index, OEvector *b);

/* Cholesky decomposition for symmetric matrices in oe_cholesky.c */
int oe_cholesky_decomposition(OEmatrix *A, OEvector *diagonal);
int oe_cholesky_substitution(OEmatrix *A, OEvector *diagonal, OEvector *b);
int oe_cholesky_invert_lower(OEmatrix *A, OEvector *diagonal);

/* Functions for printing vectors and matrices to a file in oe_print.c */
int oe_print_vector(FILE *file, OEvector *v, char *title);
int oe_print_int_vector(FILE *file, OEintvector *v, char *title);
int oe_print_row_vector(FILE *file, OEvector *v, char *title);
int oe_print_row_int_vector(FILE *file, OEintvector *v, char *title);
int oe_print_matrix(FILE *file, OEmatrix *M, char *title);
int oe_print_int_matrix(FILE *file, OEintmatrix *M, char *title);
int oe_print_vector_info(FILE *file, OEvector *v, char *title);
int oe_print_matrix_info(FILE *file, OEmatrix *M, char *title);

int oe_visualise_matrix(FILE *file, OEmatrix *M, char *title);
int oe_save_vector(char *filename, OEvector *v, char *title);
int oe_save_matrix(char *filename, OEmatrix *M, char *title);

/* Functions for creating interpolation matrices in oe_spline.c */
int oe_linear_interpolation_matrix(OEmatrix *M, int m, int n);
int oe_cubic_b_spline(OEmatrix *M, int m, int n);

/* Function for creating a Gaussian smoothing matrix in oe_gaussian.c */
int oe_gaussian_smoothing_matrix(OEmatrix *M, 
				 OEvector *range_m, OEvector *range_n,
				 OEreal drange, OEreal sigma_range);

/* Functions for compound matrix multiplication operations in oe_compound.c */
int oe_KT_invs_K(OEmatrix *K, OEvector *invS,
		 OEmatrix *sumA, OEmatrix *invS_K);
int oe_sparse_KT_invs_K(OEmatrix *K, OEvector *invS,
			OEmatrix *sumA, OEmatrix *invS_K);
int oe_sparse_W_invA_WT_diagonal(OEmatrix *W, OEmatrix *invA,
				 OEvector *v_out);

/* Return the elapsed time in oe_time.c */
OEreal oe_time_split();

/* 1D and 2D interpolation in oe_interp.c */
int oe_1d_interp(OEvector *x, OEvector *y, int ni, OEreal *xi,
		 OEreal *yi_out, OEreal *dy_dx_out);
int oe_2d_interp(OEvector *x, OEvector *y, OEmatrix *Z,
		 int ni, OEreal *xi, OEreal *yi,
		 OEreal *Zi_out, OEreal *dZ_dx_out);

/* Interface to the readconfig data structures in oe_readconfig.c */
int oe_read_vector(rc_data *data, char *name, OEvector *v);
int oe_read_matrix(rc_data *data, char *name, OEmatrix *M);

#ifdef __cplusplus
}
#endif

#endif
