/* This is rplot.h, the header file for a C++ class library for the
   fast production of graphs and images.

   Copyright (C) 2000 Robin Hogan
   <R.J.Hogan@reading.ac.uk>

   RPLOT Pros:
   + postscript, png or ppm output
   + capable handling of `raster' data
   + fonts taken from an X font server so much prettier than gnuplot!

   RPLOT Cons:
   - lots of the features you'd expect from a plotting package aren't (yet)
     implemented (contouring, 3D graphs etc).

   The `R' denotes any of `Radar', `Raster', `Reading', `Robin' etc.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef _RPLOT_H
#define _RPLOT_H

#include <list>
#include <unistd.h>
#include <string>

#ifndef FONT_CPP
typedef struct FSServer;
#endif

#define string std::string
#define ostream std::ostream
#define ofstream std::ofstream
#define ifstream std::ifstream
#define list std::list
#define setw std::setw
#define setprecision std::setprecision
#define cout std::cout
#define cerr std::cerr
#define endl std::endl

#define GRAPH_UNKNOWN -1

#define DEFAULT_TICK_LENGTH 5
#define MAX_COLORS 256
#define MAX_FONTS 10
#define MAX_CHARS 256
#define RPLOT_FILL 0 /* do we make rectangles a bit bigger to avoid white lines? */
#define RPLOT_MAX_POLYGON_VERTICES 4

//#define throw(a) cerr << (a); exit(1)

#define uchar unsigned char
#define uword unsigned short int
#define uint unsigned int

/* Useful macros */
#define MAX(a,b) (((a)>(b)) ? (a) : (b))
#define MIN(a,b) (((a)>(b)) ? (b) : (a))
#define ABS(a) ((a < 0) ? (-(a)) : (a))

/* Figure formats */
enum FORMAT {
  BITMAP_8BIT,	// Indexed image
  BITMAP_24BIT,	// RGB image
  BITMAP_32BIT,	// RGB + alpha channel
  POSTSCRIPT	// Adobe encapsulated postscript format
};

/* Text anchors */
enum ANCHOR {
  ANCHOR_CENTER,
  ANCHOR_TOP,
  ANCHOR_TOPRIGHT,
  ANCHOR_RIGHT,
  ANCHOR_BOTTOMRIGHT,
  ANCHOR_BOTTOM,
  ANCHOR_BOTTOMLEFT,
  ANCHOR_LEFT,
  ANCHOR_TOPLEFT
};

/* Line types */
enum LINE {
  LINE_NONE,
  LINE_SOLID,
  LINE_THICK,
  LINE_DOTTED
};

/* Orientations */
enum ORIENTATION {
  HORIZONTAL,
  VERTICAL
};


/* `make' function prototypes */


/* FONT CLASS */
/* Holds a basic bitmapped font */

class RFontLoader;

class RFont {
  friend class RFontLoader;
 public:
  RFont(void);
  RFont(int _pixel, int *_left, int *_right, int *_ascent, int *_descent,
	int *_width, int _max_ascent, int _max_descent,
	uint *_goffset, uint *_glength, char *_family, char *_weight, 
	char *_slant, char *_foundry, char *_xname, char *_psname, 
	uword _first, uword _last,
	uword _def, bool *_exists, uchar *_glyphs)
    : pixel(_pixel), left(_left), right(_right), ascent(_ascent),
    descent(_descent), width(_width), max_ascent(_max_ascent),
    max_descent(_max_descent), goffset(_goffset), glength(_glength),
    family(_family), weight(_weight), slant(_slant), foundry(_foundry),
    xname(_xname), psname(_psname), first(_first), last(_last),
    def(_def), exists(_exists), glyphs(_glyphs) {
    static_data = true;
  };
  ~RFont(void);
  inline bool have_glyphs();
  void show();
  int xextent(char *txt);
  int yextent(char *txt);
  int save_to_cpp_file(char *filename = NULL);
 public:
  RFontLoader *parent;
  int pixel;
  /*
  int left[MAX_CHARS], right[MAX_CHARS],
    ascent[MAX_CHARS], descent[MAX_CHARS],
    width[MAX_CHARS], max_ascent, max_descent;
  uint goffset[MAX_CHARS], glength[MAX_CHARS];
  */
  int *left, *right, *ascent, *descent, *width;
  int max_ascent, max_descent;
  uint *goffset, *glength;
  string family, weight, slant, foundry;
  string xname, psname;
  uword first, last, def;
  //  bool exists[MAX_CHARS];
  bool *exists;
  uchar *glyphs;
  bool static_data;
};

class RFontLoader {
 public:
  RFontLoader(string server_name = "", bool require_glyphs=true);
  ~RFontLoader(void);
  RFont *load(int pixels=10,
	      string family=string("helvetica"),
	      string weight=string("medium"),
	      string slant=string("r"),
	      string foundry=string("*"));
  inline RFont *load(int pixels,
		     char *family, char *weight, char *slant, char *foundry);
  int unload(RFont *font);
#ifndef NO_X_FONT_SERVER
  int open_server(string server_name=string(""));
  RFont *load_from_server(RFont *font, int pixel, string &family, 
			  string &weight, string &slant, string &foundry);
#endif
  void close_server();
  inline bool verbose();
  inline void verbose(bool v);
  inline void require_glyphs(bool rg);
 protected:
#ifndef NO_X_FONT_SERVER
  FSServer *_server;
#endif
  RFont *_font[MAX_FONTS];
#ifndef NO_X_FONT_SERVER
  string _server_name;
  pid_t _xfs_pid;
#endif
  bool _verbose, _require_glyphs;
};

class RTextElement {
  friend class RText;
  friend ostream& operator<<(ostream& os, const RTextElement &el);
 public:
  inline RTextElement(string, RFont *, int=0);
  uint width() const;
  uint height() const;
 public:
  RFont *font;
  int ascent;
  string text;
};

class RText {
  friend ostream& operator<<(ostream& os, const RText &text);
 public:
  inline RText();
  inline RText(const char *str, RFont *font);
  inline RText(string str, RFont *font);
  void render(const char *str, RFont *font);
  uint width();
  uint height();
 public:
  list<RTextElement> element;		
};


/* RECTANGLE CLASS */
/* Convenient way of passing 4 numbers around */
class RRectangle {
  friend ostream& operator<<(ostream& os, const RRectangle &rect);
 public:
  inline RRectangle();
  inline RRectangle(double xx1,double yy1,double xx2,double yy2);
  inline RRectangle operator=(const RRectangle &rect);
  // `in' methods: is (x,y) inside the rectangle?
  // `in_strict' returns false if x1>x2 or y1>y2 (rectangle has negative area)
  inline bool in(double x, double y);
  inline bool in_strict(double x, double y);
 public:
  double x1,y1,x2,y2;
};
// RRectangle `make' function
inline RRectangle make_rectangle(double x1, double y1, double x2, double y2);

/* COLOR CLASS */
class RColor {
  friend class RDrawable; 
  friend class RPalette;
  friend ostream& operator<<(ostream& os, const RColor &col);
 public:
  inline RColor(uchar r=0, uchar g=0, uchar b=0, uchar a=255);
  inline RColor(const RColor&);
  inline RColor operator=(const RColor&);
  inline bool operator==(const RColor&) const;
  inline bool operator!=(const RColor&) const;
  inline bool is_gray();
  // the `-' operator returns the squared distance in RGB space
  // between the two RColor arguments - used when a RPalette is full
  // to find the nearest color that can be used
  inline int operator-(const RColor&) const;
 public:
  uchar red, green, blue, alpha;
};
// RColor `make' function
//inline RColor make_color(uchar r, uchar g, uchar b, uchar a=255);

/* PALETTE CLASS */
/* Holds 256 colors, handles color assignment optimally */
class RPalette {
  friend class RFigure;
  friend ostream& operator<<(ostream& os, const RPalette &pal);
 public:
  inline RPalette(void);
  void clear(void);
  int load(const char *filename); // load from simple ascii file
  int write(const char *filename); // write to simple ascii file
  // Direct color assignment:
  RColor set(const uchar index, const RColor &col);
  void setrange(const int i1,const int i2,const RColor &col1,
		const RColor &col2);
  // the `+=' operator adds a color to the palette if it is not already there;
  // otherwise it returns the index of the existing color in the palette
  uchar operator+=(const RColor&);
  inline bool status(const uchar index) const; // has an entry been assigned?
  inline bool contiguous(void) const; // Are the colors contiguous from index 0?
  inline bool equal(const uchar index, const RColor &col);
  inline RColor get(const uchar index) const; // check a particular entry
  inline uchar size(void) const; // how many colors have been assigned?
  inline uchar max(void) const; // max index 
 private:
  RColor _color[MAX_COLORS];
  uchar _next, _last, _size, _max;
  bool _status[MAX_COLORS];
};

/* DRAWABLE CLASS */
/* Base class for RFigure and RAxes - do not use directly */
class RDrawable {
 public:
  inline RDrawable();
  void plot();
  void pixel(double x, double y,const RColor &);
  inline void pixel(double x, double y);
  void rectangle(double x1, double y1, double x2, double y2, const RColor &col);
  inline void rectangle(double x1, double y1, double x2, double y2);
  inline void rectangle(RRectangle &rect, const RColor &col);
  inline void rectangle(const RRectangle &rect);
  inline void clipbox(const RRectangle &rect);
  inline void clipbox(double x1, double y1, double x2, double y2);
  inline void clipbox(bool onoff=true);
  inline bool clipping();
  inline void clipbox_to_axis();
  inline void restore_clipbox();
  void text();
  inline RRectangle axis();
  inline RColor pen_color() ;
  inline RColor pen_color(const RColor &col);
  inline LINE line_type();
  inline LINE line_type(const LINE &line_type);
  inline int width();
  inline int height();
  inline double xscale();
  inline double yscale();
 protected:
  double _xscale,_yscale;
  int _width, _height;
  RRectangle _axis, _clipbox, _old_clipbox;
  RColor _pen_color;
  LINE _line_type;
  bool _clipping;
};

/* FIGURE CLASS */

class RFigure : public RDrawable {
 public:
  RFigure(int width,int height, FORMAT format, ostream *os=NULL,
	  char *title=NULL, float ps_scale = 1.0);
  RFigure(char *filename);
  ~RFigure(void);
  void blit();
  inline void axis(const RRectangle &rect);
  inline void axis(double x1, double y1, double x2, double y2);
  // Output functions: all bitmap images can be saved as PNG
  // Only 24-bit images can be saved as PPMRAW
  void write_png(char *filename);
  void write_png(ostream *);
  void write_ppmraw(char *filename);
  void write_ppmraw(ostream *os);
  inline uchar *ppmraw(); // get pointer to start of PPMRAW data
  // Drawing functions
  void clear(const RColor &col); // clear to col
  inline void clear(uchar r=255,uchar g=255,uchar b=255,uchar a=255);
  void pixel(double x, double y, const RColor &col); // draw pixel
  void plot(double x1, double y1, double x2, double y2, const RColor &col, 
	    LINE line_type=LINE_SOLID); // plot single line
  // plot line with absolute length
  void barb(double x, double y, double dx, double dy, const RColor &col,
	    LINE line_type=LINE_SOLID,
	    double *xreturn=NULL,	double *yreturn=NULL);
  void rectangle(double x1, double y1, double x2, double y2, 
		 const RColor &col); // plot solid rectangle
  void polygon(int n, float *x, float *y, const RColor &col);
  inline void rectangle(RRectangle &rect, const RColor &col); // (overloaded ver)
  RRectangle text(double x, double y, const char *str, ANCHOR anchor, 
		  int rotation, RFont *fnt, const RColor &col); // write text
  inline RRectangle text(double x, double y, const char *str, 
			 ANCHOR anchor=ANCHOR_LEFT, int rotation=0); // (overloaded ver)
  // Access functions
  inline int format();
  inline int sizeofpixel();
  inline int sizeofheader();
  inline bool is_bitmap();
 protected:
  // Protected functions for writing postscript data
  void _init_postscript(char * title, float ps_scale = 1.0);
  void _close_postscript();
  void _rectangle_postscript(float i1, float j1, float i2, float j2,
			     const RColor &col);
  void _plot_postscript(float i1, float j1, float i2, float j2, const RColor &col, 
			LINE line_type=LINE_SOLID);
  //		void _text_postscript(float x, float y, char *str, int rotation, RFont *fnt,
  //				const RColor &col);
  void _text_postscript(float x, float y, RText *text, int rotation,
			const RColor &col);
  void _color_postscript(const RColor &col);
  void _line_postscript(LINE line);
  void _polygon_postscript(int n, float *x, float *y, const RColor &col);
  // Protected functions for writing bitmap data
  void _rectangle_bitmap(int i1, int j1, int i2, int j2, const RColor &col);
  void _pixel_bitmap(int ix, int iy, const RColor &col);
  void _plot_bitmap(int i1, int j1, int i2, int j2, const RColor &col, 
		    LINE line_type=LINE_SOLID);
  //		void _text_bitmap(int x, int y, char *str, int rotation, RFont *fnt,
  //				const RColor &col);
  void _text_bitmap(int x, int y, RText *text, int rotation,
		    const RColor &col);
  void _polygon_bitmap(int n, float *x, float *y, const RColor &col);
 protected:
  float _current_x, _current_y;
  int _sizeofpixel, _sizeofheader;
  int _xoffset, _yoffset, _format;
  // Postscript data
  ostream *_ps_stream;
  RFont *_current_font;
  uchar **_data, *_pixels;
  LINE _current_line_type;
  RColor _current_color;
  bool _is_bitmap;
 public:
  // Data
  RPalette palette;
  RFont *font;
};

/* AXES CLASS */

class RAxes : public RDrawable {
 public:
  inline RAxes(RFigure *parent, const RRectangle pos);
  inline RAxes(RFigure *parent, double x1, double y1, double x2, double y2);
  inline void axis(const RRectangle &rect);
  inline void axis(double x1, double y1, double x2, double y2);
  inline RRectangle axis();
  // Drawing functions - many have same format as in RFigure
  void clear(const RColor &col); // clear to col
  inline void clear(uchar r=255,uchar g=255,uchar b=255,uchar a=255);
  void pixel(double x, double y, const RColor &col);
  inline void pixel(double x, double y);
  void plot(double x1, double y1, double x2, double y2, const RColor &col,
	    LINE line_type);
  inline void plot(double x1, double y1, double x2, double y2,
		   const RColor &col);
  inline void plot(double x1, double y1, double x2, double y2);
  inline void plot(float *x, float *y, int n, const RColor &col, LINE line_type);
  inline void plot(float *x, float *y, int n, const RColor &col);
  inline void plot(float *x, float *y, int n);
  // plot line with absolute length:
  void barb(double x, double y, double dx, double dy, const RColor &col,
	    LINE line_type, double *xreturn=NULL, double *yreturn=NULL);
  inline void barb(double x1, double y1, double dx, double dy,
		   const RColor &col);
  inline void barb(double x1, double y1, double dx, double dy);
  void rectangle(double x1, double y1, double x2, double y2, const RColor &col);
  inline void box(const RRectangle &rect,const RColor &col);
  inline void box(const RRectangle &rect);
  inline void box();
  RRectangle text(double x, double y, const char *str, ANCHOR anchor, 
		  int rotation, RFont *fnt, const RColor &col);
  inline RRectangle text(double x, double y, const char *str, 
			 ANCHOR anchor=ANCHOR_LEFT, int rotation=0,
			 RFont *fnt=NULL);
  // RAxes-specific drawing functions
  void xlabel(char *str, bool otherside=0, RFont *fnt=NULL);
  void ylabel(char *str, bool otherside=0, RFont *fnt=NULL);
  void xticks(double start, double interval, char *format=NULL, 
	      double length=DEFAULT_TICK_LENGTH, bool opposite=0, bool grid=0,
	      bool singleticks=0);
  void xticks_time(double start, double interval,
		   double length=DEFAULT_TICK_LENGTH, bool opposite=0, bool grid=0,
		   bool singleticks=0);
  void yticks(double start, double interval, char *format=NULL, 
	      double length=DEFAULT_TICK_LENGTH, bool opposite=0, bool grid=0,
	      bool singleticks=0);
  void xtick(double x, char *str, double length=DEFAULT_TICK_LENGTH, 
	     bool opposite=0, bool grid=0, bool singleticks=0);
  void ytick(double y, char *str, double length=DEFAULT_TICK_LENGTH,
	     bool opposite=0, bool grid=0, bool singleticks=0);
  void hray(float *x, float y1, float y2, float *val, int n, RPalette *pal,
	    float minval, float maxval, float nanval);
  void vray(float x1, float x2, float *y, float *val, int n, RPalette *pal,
	    float minval, float maxval, float nanval);
  void colorbar(RPalette *colormap,ORIENTATION orient=VERTICAL, int size = 0);
  void polygon(int n, float *x, float *y, const RColor &col);
  // Access/assignment functions
  inline RColor color(); // get the color of the axes (not the pen)
  inline RColor color(const RColor &col); // set the color of the axes
  inline RRectangle position();
 protected:
  RFigure *_parent;
  RRectangle _position, _extent;
  RColor _color;
  // Protected functions
  void init(RFigure *parent, const RRectangle pos); // called by constructor
};

/* OTHER FUNCTION PROTOTYPES */

void best_axis_range(double *a1, double *a2, double *da, double *na);
float best_tick_spacing(float x1, float x2, int n_ticks = 6);
void create_jet_palette(RPalette *pal,uchar num);
void create_litmus_palette(RPalette *pal,uchar num);
void create_cloud_palette(RPalette *pal,uchar num);

/* INLINE FUNCTION DEFINITIONS */

/* FONT CLASS */

inline bool RFont::have_glyphs() {
  return (glyphs != NULL);
}
/*inline RFont *load(bool require_glyphs, int pixels,
		char *fm, char *w, char *s, char *fn) {
	string sfm(fm), sw(w), ss(s), sfn(fn);
	return (load(require_glyphs, pixels, sfm, sw, ss, sfn));
*/ /*	return (load(require_glyphs, pixels, string(family), string(weight),
			string(slant), string(foundry)));*/
/*};*/

/* FONTLOADER CLASS */
inline bool RFontLoader::verbose() {
  return _verbose;
}
inline void RFontLoader::verbose(bool v) {
  _verbose = v;
}
inline void RFontLoader::require_glyphs(bool rg) {
  _require_glyphs = rg;
}

/* TEXT CLASS */
inline RText::RText() { return ; };
inline RText::RText(string str, RFont *font) { render(str.c_str(), font); };
inline RText::RText(const char *str, RFont *font) { render(str, font); };
inline RTextElement::RTextElement(string str, RFont *fnt, int asc) {
  text=str; font=fnt; ascent=asc;
}

inline RTextElement make_text_element(string str, RFont *fnt, int asc) {
  RTextElement element(str, fnt, asc);
  return element;
}

/* RECTANGLE CLASS */
inline RRectangle::RRectangle() {	x1=0;y1=0;x2=0;y2=0;	};
inline RRectangle::RRectangle(double xx1,double yy1,double xx2,double yy2) {
  x1=xx1;y1=yy1;x2=xx2;y2=yy2;
};
inline RRectangle RRectangle::operator=(const RRectangle &rect) {
  x1=rect.x1;y1=rect.y1;x2=rect.x2;y2=rect.y2;
  return rect;
};
inline bool RRectangle::in(double x, double y) {
  return ((x>x1 && x<x2) || (x>x2 && x<x1)) && 
    ((y>y1 && y<y2) || (y>y2 && y<y1));
};
inline bool RRectangle::in_strict(double x, double y) {
  return (x>x1 && x<x2 && y>y1 && y<y2);
};
// RRectangle `make' function
inline RRectangle make_rectangle(double x1, double y1, double x2, double y2) {
  RRectangle rect(x1,y1,x2,y2);
  return rect;
};

/* COLOR CLASS */
inline RColor::RColor(uchar r, uchar g, uchar b, uchar a) {
  red=r; green=g; blue=b; alpha=a;
}
inline RColor::RColor(const RColor &col) {
  red = col.red;
  green = col.green;
  blue = col.blue;
  alpha = col.alpha;
};
inline bool RColor::is_gray() {
  return ((red==green) && (red==blue));
};
inline RColor RColor::operator=(const RColor &col) {
  red=col.red;green=col.green;blue=col.blue;alpha=col.alpha;
  return col;
};
inline bool RColor::operator==(const RColor &col) const {
  return ((red==col.red) && (green==col.green) && (blue==col.blue) && 
	  (alpha==col.alpha));
}
inline bool RColor::operator!=(const RColor &col) const {
  return !((red==col.red) && (green==col.green) && (blue==col.blue) && 
	   (alpha==col.alpha));
}
inline int RColor::operator-(const RColor &col) const {
  int rdiff=(int) red - (int) col.red;
  int gdiff=(int) green - (int) col.green;
  int bdiff=(int) blue - (int) col.blue;
  return rdiff*rdiff+gdiff*gdiff+bdiff*bdiff;
};
// RColor `make' function
inline RColor make_color(uchar r, uchar g, uchar b, uchar a=255) {
  RColor col;
  col.red=r;col.green=g;col.blue=b;col.alpha=a;
  return col;
};

/* PALETTE CLASS */
inline RPalette::RPalette(void) { clear(); };
inline bool RPalette::equal(const uchar index, const RColor &col) {
  return (_color[index]==col && _status[index]);
}
inline RColor RPalette::get(const uchar index) const {	return(_color[index]); };
inline bool RPalette::status(const uchar index) const { return _status[index]; };
inline uchar RPalette::size(void) const { return _size; };
inline uchar RPalette::max(void) const { return _max; };
inline bool RPalette::contiguous(void) const { return (_max == (_size-1)); };

/* DRAWABLE CLASS */
inline	RDrawable::RDrawable() { _clipping=false; };
inline void RDrawable::pixel(double x, double y) { pixel(x,y,_pen_color); };
inline void RDrawable::rectangle(double x1, double y1, double x2, double y2) {
  rectangle(x1,y1,x2,y2,_pen_color);
};
inline void RDrawable::rectangle(RRectangle &rect, const RColor &col) {
  rectangle(rect.x1,rect.y1,rect.x2,rect.y2,col);
};
inline void RDrawable::rectangle(const RRectangle &rect) {
  rectangle(rect.x1,rect.y1,rect.x2,rect.y2,_pen_color);
};
inline void RDrawable::clipbox(const RRectangle &rect) {
  _old_clipbox=_clipbox;_clipbox=rect;_clipping=true;
};
inline void RDrawable::clipbox(double x1, double y1, double x2, double y2) {
  _old_clipbox=_clipbox;_clipbox=make_rectangle(x1,y1,x2,y2);
  _clipping=true;
};
inline void RDrawable::clipbox_to_axis() {
  _old_clipbox=_clipbox;_clipbox=_axis;_clipping=true;
};
inline void RDrawable::clipbox(bool onoff) { _clipping=onoff; };
inline bool RDrawable::clipping() { return _clipping; };
inline void RDrawable::restore_clipbox() { _clipbox=_old_clipbox;	};
inline RRectangle RDrawable::axis() { return _axis; };
inline RColor RDrawable::pen_color() { return _pen_color; };
inline RColor RDrawable::pen_color(const RColor &col) { return _pen_color=col; };
inline LINE RDrawable::line_type() { return _line_type; };
inline LINE RDrawable::line_type(const LINE &line_type) {
  return _line_type=line_type;
};
inline int RDrawable::width() { return _width; };
inline int RDrawable::height() { return _height; };
inline double RDrawable::xscale() { return _xscale; };
inline double RDrawable::yscale() { return _yscale; };

/* FIGURE CLASS */
inline void RFigure::clear(uchar r, uchar g, uchar b, uchar a) { 
  clear(make_color(r, g, b, a));
};
inline RRectangle RFigure::text(double x, double y, const char *str, 
				ANCHOR anchor, int rotation) {
  return text(x, y, str, anchor, rotation, font, _pen_color);
};
inline void RFigure::rectangle(RRectangle &rect, const RColor &col) {
  rectangle(rect.x1,rect.y1,rect.x2,rect.y2,col);
};
inline uchar *RFigure::ppmraw() {
  if (_format != BITMAP_24BIT) {
    throw string("Can only get pointer to ppmraw data for 24 bit figures");
  }
  return _pixels;
};
inline void RFigure::axis(const RRectangle &rect) {
  _axis=rect; 
  _xscale=(_axis.x2-_axis.x1)/((double) _width);
  _yscale=(_axis.y2-_axis.y1)/((double) _height);
};
inline void RFigure::axis(double x1, double y1, double x2, double y2) {
  axis(make_rectangle(x1,y1,x2,y2)); 
};
inline int RFigure::format() {	return _format; };
inline int RFigure::sizeofpixel() {	return _sizeofpixel;	};
inline int RFigure::sizeofheader() {	return _sizeofheader; };
inline bool RFigure::is_bitmap() { return _is_bitmap; };

/* AXES CLASS */
inline RAxes::RAxes(RFigure *parent, const RRectangle pos) { init(parent, pos); };
inline RAxes::RAxes(RFigure *parent, double x1, double y1, double x2, double y2) {
  init(parent,make_rectangle(x1,y1,x2,y2));
};
inline void RAxes::clear(uchar r, uchar g, uchar b, uchar a) {
  clear(make_color(r, g, b, a));
};
inline void RAxes::pixel(double x, double y) { pixel(x,y,_pen_color); };
inline void RAxes::plot(double x1, double y1, double x2, double y2,	
			const RColor &col) {
  plot(x1,y1,x2,y2,col,_line_type);
};
inline void RAxes::plot(double x1, double y1, double x2, double y2) {
  plot(x1,y1,x2,y2,_pen_color,_line_type);
};
inline void RAxes::plot(float *x, float *y, int n, const RColor &col, LINE line_type) {
  int i;
  for (i=0;i<n-1;i++) {
    plot((double) x[i], (double) y[i], (double) x[i+1], (double) y[i+1], col, 
	 line_type);
  }
};
inline void RAxes::plot(float *x, float *y, int n, const RColor &col) {
  plot(x,y,n,col,_line_type);
};
inline void RAxes::plot(float *x, float *y, int n) {
  plot(x,y,n,_pen_color,_line_type);
};
inline void RAxes::barb(double x1, double y1, double dx, double dy,
			const RColor &col) {
  barb(x1,y1,dx,dy,col,_line_type);
};
inline void RAxes::barb(double x1, double y1, double dx, double dy) {
  barb(x1,y1,dx,dy,_pen_color,_line_type);
};


inline void RAxes::box(const RRectangle &rect,const RColor &col) {
  plot(rect.x1,rect.y1,rect.x1,rect.y2,col);
  plot(rect.x1,rect.y2,rect.x2,rect.y2,col);
  plot(rect.x2,rect.y2,rect.x2,rect.y1,col);
  plot(rect.x2,rect.y1,rect.x1,rect.y1,col);
}
inline void RAxes::box(const RRectangle &rect) { box(rect,_pen_color); };
inline void RAxes::box() {  box(_axis,_color); };
inline RRectangle RAxes::text(double x, double y, const char *str, 
			      ANCHOR anchor, int rotation,
			      RFont *fnt) {
  return text(x, y, str, anchor, rotation, fnt, _pen_color);
};
inline RColor RAxes::color() { return _color; };
inline RColor RAxes::color(const RColor &col) { return _color=col; };
inline void RAxes::axis(const RRectangle &rect) {
  _axis=rect; 
  _xscale=_parent->xscale()*(_axis.x2-_axis.x1)/(_position.x2-_position.x1);
  _yscale=_parent->yscale()*(_axis.y2-_axis.y1)/(_position.y2-_position.y1);
};
inline void RAxes::axis(double x1, double y1, double x2, double y2) {
  axis(make_rectangle(x1,y1,x2,y2)); 
};
inline RRectangle RAxes::axis() { return _axis; };
inline RRectangle RAxes::position() { return _position; };

#endif
