/* oe_alloc.c -- Memory allocation functions for optimal estimation
   library

   Copyright (C) 2005 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>

#include "oe.h"

static int resize_permitted = OE_YES;
static int zero_on_reset = OE_NO;

/* Call with OE_NO as an argument to disallow any resizing (and
   associated memory allocation) of matrices and vectors. This enables
   one to be sure that parts of a program intended to be fast
   (e.g. Newtonian iteration) do not involve any memory allocation,
   and also that certain matrix multiplication functions do not
   reallocate their target matrices. Call with OE_YES to permit
   resizing. This function returns the original permission. */
int
oe_resize_permitted(int new_permit)
{
  int old_permit = resize_permitted;
  resize_permitted = new_permit;
  return old_permit;
}

/* As above but to switch the capability to automatically set to zero
   the elements of matrices and vectors whenever their size is
   reset. This is slower but it ensures that old bits of data are not
   used accidentally. */
int
oe_zero_on_reset(int new_setting)
{
  int old_setting = zero_on_reset;
  zero_on_reset = new_setting;
  return old_setting;
}

/* Reallocate a vector of reals to size n */
int
oe_reset_vector(OEvector *v, int n)
{
  if (v->n == n) {
    if (zero_on_reset) {
      OE_CHECK(oe_zero_vector(v));
    }
    return OE_SUCCESS;
  }
  else if (resize_permitted == OE_NO) {
    OE_THROW(OE_RESIZE_PERMISSION_ERROR,
	     "Attempt to resize vector of reals when resizing has been disallowed");
  }
  else if (v->_n_allocated == 0 && v->n > 0) {
    /* A linked vector: remove link before proceeding */
    v->data = NULL;
  }
  if (n > v->_n_allocated) {
    OEreal *data = oe_realloc(v->data, n*sizeof(OEreal));
    if (!data) {
      OE_THROW(OE_MEMORY_ERROR,
	       "Out of memory when allocating a vector of reals");
    }
    v->data = data;
    v->_n_allocated = v->n = n;
  }
  else {
    v->n = n;
  }
  if (zero_on_reset) {
    OE_CHECK(oe_zero_vector(v));
  }
  return OE_SUCCESS;
}

/* Reallocate a vector of ints to size n */
int
oe_reset_int_vector(OEintvector *v, int n)
{
  if (v->n == n) {
    if (zero_on_reset) {
      OE_CHECK(oe_zero_int_vector(v));
    }
    return OE_SUCCESS;
  }
  else if (resize_permitted == OE_NO) {
    OE_THROW(OE_RESIZE_PERMISSION_ERROR,
	     "Attempt to resize vector of integers when resizing has been disallowed");
  }
  else if (v->_n_allocated == 0 && v->n > 0) {
    /* A linked vector: remove link before proceeding */
    v->data = NULL;
  }
  if (n > v->_n_allocated) {
    int *data = oe_realloc(v->data, n*sizeof(int));
    if (!data) {
      OE_THROW(OE_MEMORY_ERROR,
	       "Out of memory when allocating a vector of integers");
    }
    v->data = data;
    v->_n_allocated = v->n = n;
  }
  else {
    v->n = n;
  }
  if (zero_on_reset) {
    OE_CHECK(oe_zero_int_vector(v));
  }
  return OE_SUCCESS;
}

/* Reallocate a matrix of reals to size m*n */
int
oe_reset_matrix(OEmatrix *M, int m, int n)
{
  OEreal *raw_data = NULL;
  OEreal **data = M->data;
  int j;

  if (data && M->m == m && M->n == n) {
    if (zero_on_reset) {
      OE_CHECK(oe_zero_matrix(M));
    }
    return OE_SUCCESS;
  }
  else if (resize_permitted == OE_NO) {
    OE_THROW(OE_RESIZE_PERMISSION_ERROR,
	     "Attempt to resize matrix of reals when resizing has been disallowed");
  }
  else if (data) {
    if (M->_m_allocated == 0) {
      /* A linked matrix: remove link before proceeding */
      data = NULL;
    }
    else if (M->_mn_allocated > 0) {
      raw_data = data[0];
    } /* Otherwise raw_data is external */
  }

  if (m*n > M->_mn_allocated) {
    if (!raw_data) {
      raw_data = (OEreal*) oe_malloc(m*n*sizeof(OEreal));
    }
    else {
      raw_data = (OEreal*) oe_realloc(raw_data, m*n*sizeof(OEreal));
    }
    if (!raw_data) {
      OE_THROW(OE_MEMORY_ERROR,
	       "Out of memory when allocating a matrix of reals");
    }
    M->_mn_allocated = m*n;
  }

  if (m > M->_m_allocated) {
    data = (OEreal**) oe_realloc(data, m*sizeof(OEreal*));
    if (!data) {
      OE_THROW(OE_MEMORY_ERROR,
	       "Out of memory when allocating pointers to a matrix of reals");
    }
    M->_m_allocated = m;
    data[0] = raw_data;
  }

  for (j = 0; j < m; j++) {
    data[j] = raw_data + n*j;
  }

  M->m = m;
  M->n = n;
  M->data = data;
  if (zero_on_reset) {
    OE_CHECK(oe_zero_matrix(M));
  }
  return OE_SUCCESS;
}

/* Reallocate a matrix of ints to size m*n */
int
oe_reset_int_matrix(OEintmatrix *M, int m, int n)
{
  int *raw_data = NULL;
  int **data = M->data;
  int j;

  if (data && M->m == m && M->n == n) {
    if (zero_on_reset) {
      OE_CHECK(oe_zero_int_matrix(M));
    }
    return OE_SUCCESS;
  }
  else if (resize_permitted == OE_NO) {
    OE_THROW(OE_RESIZE_PERMISSION_ERROR,
	     "Attempt to resize matrix of integers when resizing has been disallowed");
  }
  else if (data) {
    if (M->_m_allocated == 0) {
      /* A linked matrix: remove link before proceeding */
      data = NULL;
    }
    else if (M->_mn_allocated > 0) {
      raw_data = data[0];
    } /* Otherwise raw_data is external */
  }

  if (m*n > M->_mn_allocated) {
    raw_data = (int*) oe_realloc(raw_data, m*n*sizeof(int));
    if (!raw_data) {
      OE_THROW(OE_MEMORY_ERROR,
	       "Out of memory when allocating a matrix of integers");
    }
    M->_mn_allocated = m*n;
  }

  if (m > M->_m_allocated) {
    data = (int**) oe_realloc(data, m*sizeof(int*));
    if (!data) {
      OE_THROW(OE_MEMORY_ERROR,
	 "Out of memory when allocating pointers to a matrix of integers");
    }
    M->_m_allocated = m;
    data[0] = raw_data;
  }

  for (j = 0; j < m; j++) {
    data[j] = raw_data + n*j;
  }

  M->m = m;
  M->n = n;
  M->data = data;
  if (zero_on_reset) {
    OE_CHECK(oe_zero_int_matrix(M));
  }
  return OE_SUCCESS;
}

/* Create a matrix Mdest whose rows point to rows in matrix Msrc (as
   determined by the contents of irows) but spanning only columns from
   start_col to end_col in Msrc. */
int
oe_subset_matrix(OEmatrix *Mdest, OEmatrix *Msrc, OEintvector *irows,
		 int start_col, int end_col)
{
  int j_dest;
  int m = Msrc->m, n = Msrc->n;
  int m_dest = irows->n, n_dest = end_col-start_col+1;
  if (resize_permitted == OE_NO) {
    OE_THROW(OE_RESIZE_PERMISSION_ERROR,
	     "Cannot subset a matrix when resizing has been explicitly disallowed");
  }
  if (m < irows->data[m_dest-1]) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Source matrix to be subset has too few rows");
  }
  if (n < end_col) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Source matrix to be subset has too few columns");
  }
  if (start_col > end_col) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Start column is greater than end column in subset matrix");
  }
  
  oe_free_matrix(Mdest);
  Mdest->data = (OEreal**) oe_malloc(m_dest*sizeof(OEreal*));
  if (!Mdest->data) {
    OE_THROW(OE_MEMORY_ERROR,
	     "Out of memory when allocating pointers to subset matrix");
  }
  Mdest->_m_allocated = m_dest;
  Mdest->_mn_allocated = 0;
  Mdest->m = m_dest;
  Mdest->n = n_dest;
  for (j_dest = 0; j_dest < m_dest; j_dest++) {
    Mdest->data[j_dest] = Msrc->data[irows->data[j_dest]] + start_col;
  }
  return OE_SUCCESS;
}

/* Create a matrix Mdest whose rows point to rows in matrix Msrc from
   start_row to end_row spanning only columns from start_col to
   end_col in Msrc. */
int
oe_block_subset_matrix(OEmatrix *Mdest, OEmatrix *Msrc, 
		       int start_row, int end_row,
		       int start_col, int end_col)
{
  int j_dest;
  int m = Msrc->m, n = Msrc->n;
  int m_dest = end_row-start_row+1, n_dest = end_col-start_col+1;
  if (resize_permitted == OE_NO) {
    OE_THROW(OE_RESIZE_PERMISSION_ERROR,
	     "Cannot block subset a matrix when resizing has been explicitly disallowed");
  }
  if (m < end_row) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Source matrix to be subset has too few rows");
  }
  if (n < end_col) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Source matrix to be subset has too few columns");
  }
  if (start_row > end_row) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Start row is greater than end row in subset matrix");
  }
  if (start_col > end_col) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Start column is greater than end column in subset matrix");
  }
  
  oe_free_matrix(Mdest);
  Mdest->data = (OEreal**) oe_malloc(m_dest*sizeof(OEreal*));
  if (!Mdest->data) {
    OE_THROW(OE_MEMORY_ERROR,
	     "Out of memory when allocating pointers to subset matrix");
  }
  Mdest->_m_allocated = m_dest;
  Mdest->_mn_allocated = 0;
  Mdest->m = m_dest;
  Mdest->n = n_dest;
  for (j_dest = 0; j_dest < m_dest; j_dest++) {
    Mdest->data[j_dest] = Msrc->data[start_row + j_dest] + start_col;
  }
  return OE_SUCCESS;
}

/* Create a vector that points to external data */
int
oe_link_vector(OEvector *v, int n, OEreal *data)
{
  oe_free_vector(v);
  v->data = data;
  v->n = n;
  v->_n_allocated = 0;
  return OE_SUCCESS;
}

/* Create a matrix that points to external data */
int
oe_link_matrix(OEmatrix *M, int m, int n, OEreal **data)
{
  oe_free_matrix(M);
  M->data = data;
  M->m = m;
  M->n = n;
  M->_mn_allocated = 0;
  M->_m_allocated = 0;
  return OE_SUCCESS;
}

/* Free the memory allocated to a vector of reals */
void
oe_free_vector(OEvector *v)
{
  if (v->data && v->_n_allocated) {
    oe_free(v->data);
  }
  v->n = 0;
  v->_n_allocated = 0;
  v->data = NULL;
}

/* Free the memory allocated to a vector of integers */
void
oe_free_int_vector(OEintvector *v)
{
  if (v->data && v->_n_allocated) {
    oe_free(v->data);
  }
  v->n = 0;
  v->_n_allocated = 0;
  v->data = NULL;
}

/* Free the memory allocated to a matrix of reals */
void
oe_free_matrix(OEmatrix *M)
{
  if (M->data) {
    if (M->_mn_allocated) {
      oe_free(M->data[0]);
    }
    if (M->_m_allocated) {
      oe_free(M->data);
    }
  }
  M->m = M->n = 0;
  M->_mn_allocated = 0;
  M->_m_allocated = 0;
  M->data = NULL;
}

/* Free the memory allocated to a matrix of integers */
void
oe_free_int_matrix(OEintmatrix *M)
{
  if (M->data) {
    if (M->_mn_allocated) {
      oe_free(M->data[0]);
    }
    if (M->_m_allocated) {
      oe_free(M->data);
    }
  }
  M->m = M->n = 0;
  M->_mn_allocated = 0;
  M->_m_allocated = 0;
  M->data = NULL;
}
