/* oe_multiply.c -- Matrix multiplication functions for optimal
   estimation library

   Copyright (C) 2005 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "oe.h"

/* Multiply matrix M with column vector v and return result in vector
   v_out */
int
oe_matrix_x_vector(OEmatrix *M, OEvector *v, OEvector *v_out)
{
  int m = M->m;
  int n = M->n;
  int i, j;

  if (n != v->n) {
    OE_THROW(OE_DIMENSION_ERROR, 
	     "Vector length must equal number of columns in matrix");
  }
  
  OE_CHECK(oe_reset_vector(v_out, m));
  
  for (j = 0; j < m; j++) {
    v_out->data[j] = 0.0;
    for (i = 0; i < n; i++) {
      v_out->data[j] += M->data[j][i] * v->data[i];
    }
  }
  return OE_SUCCESS;
}

/* Multiply the transpose of matrix M with column vector v and return
   result in vector v_out */
int
oe_matrixT_x_vector(OEmatrix *M, OEvector *v, OEvector *v_out)
{
  int m = M->m;
  int n = M->n;
  int i, j;

  if (m != v->n) {
    OE_THROW(OE_DIMENSION_ERROR, 
	     "Vector length must equal number of rows in matrix");
  }
  
  OE_CHECK(oe_reset_vector(v_out, n));
  OE_CHECK(oe_zero_vector(v_out));

  for (j = 0; j < m; j++) {
    for (i = 0; i < n; i++) {
      v_out->data[i] += M->data[j][i] * v->data[j];
    }
  }
  return OE_SUCCESS;
}

/* Interpret vector v as a diagonal matrix, and multiply by M, giving
   the result in M_out */
int
oe_diagonal_x_matrix(OEvector *v, OEmatrix *M, OEmatrix *M_out)
{
  int m = M->m;
  int n = M->n;
  int i, j;

  if (m != v->n) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Vector length must equal number of columns in matrix");
  }
  
  OE_CHECK(oe_reset_matrix(M_out, m, n));
  
  for (j = 0; j < m; j++) {
    for (i = 0; i < n; i++) {
      M_out->data[j][i] = M->data[j][i] * v->data[j];
    }
  }
  return OE_SUCCESS;
}

/* Add matrix "src" to matrix "target" */
int
oe_add_matrix(OEmatrix *src, OEmatrix *target)
{
  int m = target->m;
  int n = target->n;
  int i, j;
  if (n != src->n || m != src->m) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Matrices must be the same size");
  }
  for (j = 0; j < m; j++) {
    for (i = 0; i < n; i++) {
      target->data[j][i] += src->data[j][i];
    }
  }
  return OE_SUCCESS;
}

/* Subtract matrix "src" from matrix "target" */
int
oe_subtract_matrix(OEmatrix *src, OEmatrix *target)
{
  int m = target->m;
  int n = target->n;
  int i, j;
  if (n != src->n || m != src->m) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Matrices must be the same size");
  }
  for (j = 0; j < m; j++) {
    for (i = 0; i < n; i++) {
      target->data[j][i] -= src->data[j][i];
    }
  }
  return OE_SUCCESS;
}

/* Add matrix "banddiag" to matrix "target", but interpretting
   banddiag as a band-diagonal matrix with non-zero elements only up
   to "off_diagonals" away from the diagonal */
int
oe_add_band_diagonal(OEmatrix *banddiag, int off_diagonals, OEmatrix *target)
{
  int n = target->n;
  int i, j;
  if (n != target->m || n != banddiag->n || n != banddiag->m) {
    OE_THROW(OE_DIMENSION_ERROR, 
	     "Matrices must be square and the same size");
  }
  if (off_diagonals < 0) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Cannot have a negative number of off_diagonals");
  }
  for (j = 0; j < n; j++) {
    int imax = (j+off_diagonals<n-1) ? j+off_diagonals : n-1;
    i = (j>off_diagonals) ? (j-off_diagonals) : 0;
    for ( ; i <= imax; i++) {
      target->data[j][i] += banddiag->data[j][i];
    }
  }
  return OE_SUCCESS;
}


/* Multiply band_diagonal matrix M with column vector v and return
   result in vector v_out */
int
oe_band_diagonal_x_vector(OEmatrix *M, OEvector *v, OEvector *v_out)
{
  int m = M->m;
  int n = M->n;
  int i, j, istart = 0, istart_new;

  if (n != v->n) {
    OE_THROW(OE_DIMENSION_ERROR, 
	     "Vector length must equal number of columns in matrix");
  }
  
  OE_CHECK(oe_reset_vector(v_out, m));

#define ROWS_THEN_COLUMNS
#ifdef ROWS_THEN_COLUMNS
  for (j = 0; j < m; j++) {
    v_out->data[j] = 0.0;
    istart_new = istart;
    while (istart_new < n && M->data[j][istart_new] == 0.0) {
      istart_new++;
    }
    if (istart_new < n) {
      i = istart = istart_new;
      do {
	v_out->data[j] += M->data[j][i] * v->data[i];
	i++;
      } while (i < n && M->data[j][i] != 0.0);
    }
  }
#else
  OE_CHECK(oe_zero_vector(v_out));
  for (i = 0; i < n; i++) {
    istart_new = istart;
    while (istart_new < m && M->data[istart_new][i] == 0.0) {
      istart_new++;
    }
    if (istart_new < m) {
      j = istart = istart_new;
      do {
	v_out->data[j] += M->data[j][i] * v->data[i];
	j++;
      } while (j < m && M->data[j][i] != 0.0);
    }
  }
#endif
  return OE_SUCCESS;
}

/* Interpret vector v as a diagonal matrix, and multiply by the band
   diagonal matrix M, giving the result in M_out */
int
oe_diagonal_x_band_diagonal(OEvector *v, OEmatrix *M, OEmatrix *M_out)
{
  int m = M->m;
  int n = M->n;
  int i, j, istart = 0;

  if (m != v->n) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Vector length must equal number of columns in matrix");
  }
  
  OE_CHECK(oe_reset_matrix(M_out, m, n));
  OE_CHECK(oe_zero_matrix(M_out));

  for (j = 0; j < m; j++) {
    while (M->data[j][istart] == 0.0) {
      istart++;
    }
    i = istart;
    do {
      M_out->data[j][i] = M->data[j][i] * v->data[j];
    } while (++i < n && M->data[j][i] != 0.0);
  }
  return OE_SUCCESS;
}

/* Non-sparse O(N^3) matrix multiplication - avoid using this if you
   can! */
int
oe_matrix_x_matrix(OEmatrix *M1, OEmatrix *M2, OEmatrix *M_out)
{
  int i, j, h;
  int m = M1->m;
  int n = M2->n;
  int l = M1->n;

  if (l != M2->m) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Inner dimensions disagree in matrix multiplication");
  }

  OE_CHECK(oe_reset_matrix(M_out, m, n));

  for (j = 0; j < m; j++) {
    for (i = 0; i < n; i++) {
      M_out->data[j][i] = 0.0;
      for (h = 0; h < l; h++) {
	M_out->data[j][i] += M1->data[j][h] * M2->data[h][i];
      }
    }
  }
  return OE_SUCCESS;
}



/* Multiply band_diagonal matrix M1 and matrix M2 together and return
   result in M_out */
int
oe_band_diagonal_x_matrix(OEmatrix *M1, OEmatrix *M2, OEmatrix *M_out)
{
  int i, j, h;
  int m = M1->m;
  int n = M2->n;
  int l = M1->n;

  int hstart = 0, hstart_new, hend;

  if (l != M2->m) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Inner dimensions disagree in matrix multiplication");
  }
  
  OE_CHECK(oe_reset_matrix(M_out, m, n));

  for (j = 0; j < m; j++) {
    /* For row j of M1, find non-zero elements */
    hstart_new = hstart;
    while (hstart_new < l && M1->data[j][hstart_new] == 0.0) {
      hstart_new++;
    }
    if (hstart_new >= l) {
      /* A row of zeros only */
      for (i = 0; i < n; i++) {
	M_out->data[j][i] = 0.0;
      }
    }
    else {
      hstart = hstart_new;
      /* Find final non-zero point */
      hend = hstart+1;
      while (hend < l && M1->data[j][hend] != 0.0) {
	hend++;
      }
      hend--;
      for (i = 0; i < n; i++) {
	M_out->data[j][i] = 0.0;
	for (h = hstart; h <= hend; h++) {
	  M_out->data[j][i] += M1->data[j][h] * M2->data[h][i];
	}
      }
    }
  }
  return OE_SUCCESS;
}


/* Multiply matrix M1 and band-diagonal matrix M2 together and return
   result in M_out */
int
oe_matrix_x_band_diagonal(OEmatrix *M1, OEmatrix *M2, OEmatrix *M_out)
{
  int i, j, h;
  int m = M1->m;
  int n = M2->n;
  int l = M1->n;

  int hstart = 0, hstart_new, hend;

  if (l != M2->m) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Inner dimensions disagree in matrix multiplication");
  }
  
  OE_CHECK(oe_reset_matrix(M_out, m, n));

  for (i = 0; i < n; i++) {
    /* For column i of M2, find non-zero elements */
    hstart_new = hstart;
    while (hstart_new < l && M2->data[hstart_new][i] == 0.0) {
      hstart_new++;
    }
    if (hstart_new >= l) {
      /* A column of zeros only */
      for (j = 0; j < m; j++) {
	M_out->data[j][i] = 0.0;
      }
    }
    else {
      hstart = hstart_new;
      /* Find final non-zero point */
      hend = hstart+1;
      while (hend < l && M2->data[hend][i] != 0.0) {
	hend++;
      }
      hend--;
      for (j = 0; j < m; j++) {
	M_out->data[j][i] = 0.0;
	for (h = hstart; h <= hend; h++) {
	  M_out->data[j][i] += M1->data[j][h] * M2->data[h][i];
	}
      }
    }
  }
  return OE_SUCCESS;
}
