/* oe_print.h -- Print matrices and vectors in optimal estimation
   library

   Copyright (C) 2005 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <math.h>

#include "oe.h"

/* Print vector "v" to "file", heading it by the optional "title". */
int
oe_print_vector(FILE *file, OEvector *v, char *title)
{
  int i;
  if (title) {
    fprintf(file, "%s (%dx1 vector, %d allocated)\n",
	    title, v->n, v->_n_allocated);
  }
  for (i = 0; i < v->n; i++) {
    fprintf(file, "\t%g\n", v->data[i]);
  }
  return OE_SUCCESS;
}

/* Print info about vector "v" to "file", heading it by the optional
   "title". */
int
oe_print_vector_info(FILE *file, OEvector *v, char *title)
{
  if (title) {
    fprintf(file, "%s: %dx1 vector, %d allocated, missing_value=%g, location=%p\n",
	    title, v->n, v->_n_allocated, v->missing_value, v->data);
  }
  return OE_SUCCESS;
}

/* Print integer vector "v" to "file", heading it by the optional
   "title". */
int
oe_print_int_vector(FILE *file, OEintvector *v, char *title)
{
  int i;
  if (title) {
    fprintf(file, "%s (%dx1 integer vector, %d allocated)\n",
	    title, v->n, v->_n_allocated);
  }
  for (i = 0; i < v->n; i++) {
    fprintf(file, "\t%d\n", (int) v->data[i]);
  }
  return OE_SUCCESS;
}

/* Print row vector "v" to "file", heading it by the optional
   "title". */
int
oe_print_row_vector(FILE *file, OEvector *v, char *title)
{
  int i;
  if (title) {
    fprintf(file, "%s (%dx1 vector, %d allocated)\n",
	    title, v->n, v->_n_allocated);
  }
  for (i = 0; i < v->n; i++) {
    fprintf(file, "\t%g", v->data[i]);
  }
  fprintf(file, "\n");
  return OE_SUCCESS;
}

/* Print integer row vector "v" to "file", heading it by the optional
   "title". */
int
oe_print_row_int_vector(FILE *file, OEintvector *v, char *title)
{
  int i;
  if (title) {
    fprintf(file, "%s (%dx1 integer vector, %d allocated)\n",
	    title, v->n, v->_n_allocated);
  }
  for (i = 0; i < v->n; i++) {
    fprintf(file, "\t%d", (int) v->data[i]);
  }
  fprintf(file, "\n");
  return OE_SUCCESS;
}

/* Print matrix "M" to "file", heading it by the optional "title". */
int
oe_print_matrix(FILE *file, OEmatrix *M, char *title)
{
  int i, j;
  if (title) {
    fprintf(file, "%s (%dx%d matrix, %d allocated)\n",
	    title, M->m, M->n, M->_mn_allocated);
  }
  for (j = 0; j < M->m; j++) {
    for (i = 0; i < M->n; i++) {
      fprintf(file, " %10.7g", M->data[j][i]);
    }
    fprintf(file, "\n");
  }
  return OE_SUCCESS;
}

/* Print integer matrix "M" to "file", heading it by the optional
   "title". */
int
oe_print_int_matrix(FILE *file, OEintmatrix *M, char *title)
{
  int i, j;
  if (title) {
    fprintf(file, "%s (%dx%d integer matrix, %d allocated)\n",
	    title, M->m, M->n, M->_mn_allocated);
  }
  for (j = 0; j < M->m; j++) {
    for (i = 0; i < M->n; i++) {
      fprintf(file, " %d", M->data[j][i]);
    }
    fprintf(file, "\n");
  }
  return OE_SUCCESS;
}

/* Print an "ASCII-Art" version of a matrix to "file". */
int
oe_visualise_matrix(FILE *file, OEmatrix *M, char *title)
{
  int i, j, inf = 0;
  if (title) {
    fprintf(file, "%s (%dx%d matrix, %d allocated): "
	    "0=zero, +=positive, -=negative, *=missing, N=NaN, i=-Inf, I=+Inf\n",
	    title, M->m, M->n, M->_mn_allocated);
  }
  for (j = 0; j < M->m; j++) {
    for (i = 0; i < M->n; i++) {
      if (isnan(M->data[j][i])) {
	fprintf(file, "N");	
      }
      else if ((inf = isinf(M->data[j][i])) != 0) {
	if (inf < 0 ) {
	  fprintf(file, "i");
	}
	else {
	  fprintf(file, "I");
	}
      }
      else if (M->data[j][i] == 0.0) {
	fprintf(file, "0");
      }
      else if (M->data[j][i] == M->missing_value) {
	fprintf(file, "*");
      }
      else if (M->data[j][i] < 0.0) {
	fprintf(file, "-");
      }
      else {
	fprintf(file, "+");
      }
    }
    fprintf(file, "\n");
  }
  return OE_SUCCESS;
}

/* Print info about matrix "M" to "file", heading it by the optional
   "title". */
int
oe_print_matrix_info(FILE *file, OEmatrix *M, char *title)
{
  if (title) {
    fprintf(file, "%s: %dx%d matrix, %d allocated, missing_value=%g, location=%p\n",
	    title, M->m, M->n, M->_mn_allocated, M->missing_value, M->data);
  }
  return OE_SUCCESS;
}

/* Save vector as ASCII to a file named "filename", with first line
   given by "title". */ 
int
oe_save_vector(char *filename, OEvector *v, char *title)
{
  FILE *outfile = fopen(filename, "w");
  if (outfile) {
    OE_CHECK(oe_print_vector(outfile, v, title));
    fclose(outfile);
  }
  else {
    OE_THROW(OE_FAILURE, "Failed to open text file into which to write vector");
  }
  return OE_SUCCESS;
}

/* Save matrix as ASCII to a file named "filename", with first line
   given by "title". */ 
int
oe_save_matrix(char *filename, OEmatrix *M, char *title)
{
  FILE *outfile = fopen(filename, "w");
  if (outfile) {
    OE_CHECK(oe_print_matrix(outfile, M, title));
    fclose(outfile);
  }
  else {
    OE_THROW(OE_FAILURE, "Failed to open text file into which to write matrix");
  }
  return OE_SUCCESS;
}



