/* oe_solve.h -- Functions for solving linear equations in the optimal
   estimation library

   Copyright (C) 2005 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include "oe.h"

/* Internal data */
static OEvector _diagonal = OE_NEW_VECTOR;
static OEintvector _index = OE_NEW_VECTOR; 

/* Solve the system of equations A*x=b, overwriting b with the result
   x. This function uses LU decomposition. If A is symmetric then use
   oe_solve_symmetric() instead.  */
int
oe_solve(OEmatrix *A, OEvector *b)
{
  int interchanges_even;
  OE_CHECK(oe_reset_int_vector(&_index, b->n));
  OE_CHECK(oe_lu_decomposition(A, _index.data, &interchanges_even));
  OE_CHECK(oe_lu_substitution(A, _index.data, b));
  return OE_SUCCESS;
}

/* Invert matrix A and return the result in invA. A will be
   corrupted. This function uses LU decomposition. If A is symmetric,
   use the function oe_invert_symmetric(). */
int
oe_invert(OEmatrix *A, OEmatrix *invA)
{
  int j;
  int interchanges_even;

  OE_CHECK(oe_reset_int_vector(&_index, A->n));
  OE_CHECK(oe_zero_matrix(invA));
  OE_CHECK(oe_lu_decomposition(A, _index.data, &interchanges_even));

  for (j = 0; j < A->n; j++) {
    OEvector column = OE_EXTRACT_ROW(invA, j);
    column.data[j] = 1.0;
    OE_CHECK(oe_lu_substitution(A, _index.data, &column));
  }
  return OE_SUCCESS;
}

/* Solve the set of equations A*x=b for symmetric matrix A,
   overwriting b with the result. Cholesky decomposition is used. A
   will be corrupted. */
int
oe_solve_symmetric(OEmatrix *A, OEvector *b)
{
  OE_CHECK(oe_cholesky_decomposition(A, &_diagonal));
  OE_CHECK(oe_cholesky_substitution(A, &_diagonal, b));
  return OE_SUCCESS;
}

/* Invert symmetric matrix A and overwrite A with the result. */
int
oe_invert_symmetric(OEmatrix *A, OEmatrix *invA)
{
  int j;

  OE_CHECK(oe_zero_matrix(invA));
  OE_CHECK(oe_cholesky_decomposition(A, &_diagonal));

  for (j = 0; j < A->n; j++) {
    OEvector column = OE_EXTRACT_ROW(invA, j);
    column.data[j] = 1.0;
    OE_CHECK(oe_cholesky_substitution(A, &_diagonal, &column));
  }

  return OE_SUCCESS;
}
