/* 
   Copyright (C) 2000 Robin Hogan

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iomanip>
#include <time.h>
#include "rplot.h"

#define PS_SCALE 10.0
#define SCALE(a) ((int) (PS_SCALE * (a)))

void
RFigure::_init_postscript(char *title, float scale)
{
  _xoffset = (int) (298.0/scale - 0.5 * _width);
  _yoffset = (int) (371.0/scale - 0.5 * _height);
  _current_font=NULL;
  _current_x=-1000, _current_y=-1000;
  _current_line_type=_line_type;
  _current_color=_pen_color;

  if (scale <= 0) {
    scale = 1.0;
  }
  time_t t=time(NULL);
	

  /* Postscript header */
  (*_ps_stream) <<
    "%!PS-Adobe-3.0 EPSF-3.0\n"
    "%%Creator: LibRPlot\n";
  if (title != NULL) {
    (*_ps_stream) << "%%Title: " << title << "\n";
  }
  (*_ps_stream) << "%%CreationDate: " << ctime(&t) << 
    "%%DocumentNeededFonts: Helvetica\n"
    "%%LanguageLevel: 2\n"
    "%%Pages: 1\n"
    "%%BoundingBox:    " << 
    (int) (scale * _xoffset) << " " << (int) (scale * _yoffset) << " " << 
    (int) (scale * (_width+_xoffset)) << " " 
		<< (int) (scale * (_height+_yoffset)) << "\n"
    "%%EndComments\n";

  /* Postscript prolog */
  (*_ps_stream) <<
    "%%BeginProlog\n"
    "\n"
    "% LibRPlot dictionary\n"
    "/LibRPlot 150 dict begin\n"
    "\n"
    "/bd {bind def} bind def\n"
    "/ld {load def} bind def\n"
    "/b {rectclip} bd\n"
    "/r {rectfill} bd\n"
    "/m /moveto ld\n"
    "/c /setrgbcolor ld\n"
    "/g /setgray ld\n"
    "/s /show ld\n"
    "/w /setlinewidth ld\n"
    "/cap /setlinecap ld\n"
    "/pagesave () def\n"
    "/bp {/pagesave save def} bd\n"
    "/ep {pagesave restore} bd\n"
    "/rot /rotate ld\n"
    "/Fs 0 def\n"
    "/Gs 0 def\n"
    "\n"
    "% polygons\n"
    "/p {\n"
    "	3 1 roll moveto\n"
    "	1 sub {rlineto} repeat\n"
    "	closepath eofill\n"
    "} bd\n"
    "\n"
    "% font control\n"
    "/f {\n"
    "	/Gs exch store\n"
    "	exch /ISOLatin1Encoding 1 index dup \n"
    "	findfont /CharStrings get /Agrave known\n"
    "	{ \n"
    "		exch dup where\n"
    "		{pop load} {pop StandardEncoding} ifelse\n"
    "		exch dup 3 1 roll\n"
    "		findfont dup length dict begin\n"
    "			{ 1 index /FID ne {def}{pop pop} ifelse } forall\n"
    "			/Encoding exch def\n"
    "  			currentdict\n"
    "		end\n"
    "		definefont pop\n"
    "	}\n"
    "	{pop pop} ifelse\n"
    "	findfont exch /Fs exch store\n"
    "	[Fs 0 0 Fs 0 Gs] makefont setfont\n"
    "} bd\n"
    "\n"
    "% macros for lines and objects\n"
    "/L { lineto stroke } bd\n"
    "/S {" << PS_SCALE << " w [] 0 setdash } bd\n"
    "/T {" << PS_SCALE*2.5 << " w [] 0 setdash } bd\n"
    "/D {" << PS_SCALE << " w [" << PS_SCALE <<  " " << PS_SCALE*4 
    		<< "] 0 setdash } bd\n"
    "\n"
    "currentdict end def\n"
    "%%EndProlog\n";

  /* Postscript setup */
  (*_ps_stream) <<
    "%%BeginSetup\n"
    "LibRPlot begin\n"
    "0 cap\n" << (float) scale/PS_SCALE << " " << (float) scale/PS_SCALE << " scale\n"
    "S\n"
    "end\n"
    "%%EndSetup\n"
    "\n"
    "%%Page: 1 1\n"
    "%%BeginPageSetup\n"
    "%%PageBoundingBox:    " << 
    (int) (scale * _xoffset) << " " << (int) (scale * _yoffset) << " " << 
    (int) (scale * (_width+_xoffset)) << " " 
		<< (int) (scale * (_height+_yoffset)) << "\n"
    "LibRPlot begin\n"
    "bp\n"
    "%%EndPageSetup\n";

  return;
}

void
RFigure::_close_postscript()
{

  (*_ps_stream) <<
    "\n"
    "ep\n"
    "end\n"
    "\n"
    "showpage\n"
    "\n"
    "%%Trailer\n"
    "%%EOF\n";
  _ps_stream->flush();
  return;
}

void
RFigure::_color_postscript(const RColor &col)
{
  if (col != _current_color) {
    _current_color=col;
    if (_current_color.is_gray()) {
      (*_ps_stream) << setprecision(3) << (float) col.red/255 << 
	setprecision(6) << " g\n";
    }
    else {
      (*_ps_stream) << setprecision(3) << (float) col.red/255 << " " << 
	(float) col.green/255 << " " <<	(float) col.blue/255 << 
	setprecision(6) << " c\n";
    }
  }
  return;
}

void
RFigure::_line_postscript(LINE line)
{
  if (line != _current_line_type) {
    _current_line_type=line;
    switch (line) {
    case LINE_SOLID:
      (*_ps_stream) << "S\n";
      break;
    case LINE_THICK:
      (*_ps_stream) << "T\n";
      break;
    case LINE_DOTTED:
      (*_ps_stream) << "D\n";
      break;
    default:
      break;
    }
  }
  return;
}

void
RFigure::_rectangle_postscript(float i1, float j1, float i2, float j2,
				    const RColor &col)
{
  _color_postscript(col);
  (*_ps_stream) << SCALE(i1) << " " << SCALE(j1) << 
    " " << SCALE(i2-i1) << " " << SCALE(j2-j1) <<
    " r\n";
  return;
}

void
RFigure::_plot_postscript(float i1, float j1, float i2, float j2, 
			  const RColor &col, LINE line_type)
{
  _color_postscript(col);
  _line_postscript(line_type);
  (*_ps_stream) << SCALE(i1) << " " << SCALE(j1) <<
    " m " << SCALE(i2) << " " << SCALE(j2) << " L\n";
  return;
}

void
RFigure::_text_postscript(float x, float y, RText *text, int rotation, 
			  const RColor &col)
{
  unsigned int ch = 0;
  float fudged_ascent = -1;
  list<RTextElement>::const_iterator i=text->element.begin();
  _color_postscript(col);

  (*_ps_stream) << SCALE(x) << " " << SCALE(y) << " m\n"; 
  if (rotation != 0) {
    (*_ps_stream) << rotation << " rot\n";
  }

  for ( ; i != text->element.end(); i++) {

    if (i->font != _current_font) {
      _current_font=i->font;
      if (fudged_ascent < 0) {
	fudged_ascent = _current_font->pixel * 0.25;
      }
      (*_ps_stream) << "/" << _current_font->psname << " " 
		    << PS_SCALE * _current_font->pixel << " " 
		    << PS_SCALE * (i->ascent + fudged_ascent) << " f\n";
    }
    (*_ps_stream) << "(";
    for (ch=0; ch<(i->text.size()); ch++) {
      if (i->text[ch] == '(' || i->text[ch] == ')') {
	(*_ps_stream) << '\\';
      }
      (*_ps_stream) << i->text[ch];
    }
    (*_ps_stream) << ") s\n";

  }

  if (rotation != 0) {
    (*_ps_stream) << -rotation << " rot\n";
  }
  return;
}

void
RFigure::_polygon_postscript(int n, float *x, float *y, const RColor &col)
{
  int i;
  _color_postscript(col);
  if (n < 3) {
    return;
  }
  for (i = 1; i < n; ++i) {
    (*_ps_stream) << SCALE(x[i-1]-x[i]) << " " << SCALE(y[i-1]-y[i]) << " ";
  }
  (*_ps_stream) << SCALE(x[n-1]) << " " << SCALE(y[n-1]) << " " << n << " p\n";
  return;
}
