#include <math.h>
#include "zdr_phidp.h"

/* Foote and Dutoit (1969) formula for terminal fall speed at mean sea level 
 * D is in mm, vt in m/s
 */
static
OEreal
foote_dutoit(OEreal diameter)
{
  OEreal vt;
  if (diameter > 6.0) {
    diameter = 6.0;
  }
  vt =
    -0.193
    + 4.96*diameter 
    - 0.904*diameter*diameter
    + 0.0566*diameter*diameter*diameter;
  return vt;
}

void
free_scattering_lut(ScatteringLUT *lut)
{
  oe_free_matrix(&lut->ln_Zh_R);
  oe_free_matrix(&lut->ln_Zh_Kdp);
  oe_free_matrix(&lut->ln_Zh_attenh);
  oe_free_matrix(&lut->ln_Zh_attenv);
  oe_free_matrix(&lut->Zdr);
  oe_free_vector(&lut->temperature);
  oe_free_vector(&lut->D0);
}

int
create_scattering_lut(ScatteringData *scattering,
		      int nD0, OEreal D0start, OEreal dD0,
		      OEreal mu, ScatteringLUT *lut)
{
  int nT = scattering->temperature.n;
  int nD = scattering->diameter.n;
  int iD0, iD, iT;

  OEvector terminal_velocity = OE_NEW_VECTOR;
  OEvector volume = OE_NEW_VECTOR;

  /* Reset elements of lut structure */
  OE_CHECK(oe_reset_matrix(&lut->ln_Zh_R, nT, nD0));
  OE_CHECK(oe_reset_matrix(&lut->ln_Zh_Kdp, nT, nD0));
  OE_CHECK(oe_reset_matrix(&lut->ln_Zh_attenh, nT, nD0));
  OE_CHECK(oe_reset_matrix(&lut->ln_Zh_attenv, nT, nD0));
  OE_CHECK(oe_reset_matrix(&lut->Zdr, nT, nD0));
  OE_CHECK(oe_copy_vector(&scattering->temperature, &lut->temperature));
  OE_CHECK(oe_reset_vector(&lut->D0, nD0));
  lut->mu = mu;
  
  OE_CHECK(oe_reset_vector(&terminal_velocity, nD));
  OE_CHECK(oe_reset_vector(&volume, nD));

  /* Calculate the drop fall speeds */
  for (iD = 0; iD < nD; iD++) {
    terminal_velocity.data[iD] = foote_dutoit(scattering->diameter.data[iD]);
    volume.data[iD] = (scattering->diameter.data[iD]
		       *scattering->diameter.data[iD]
		       *scattering->diameter.data[iD])
      * M_PI * 1.0e-9 / 6.0;
  }

  /* Create lookup table */
  /* Loop over median volumetric diameter */
  for (iD0 = 0; iD0 < nD0; iD0++) {
    /* Set the median volumetric diameter */
    lut->D0.data[iD0] = D0start + iD0*dD0;

    /* Loop over temperature */
    for (iT = 0; iT < nT; iT++) {
      OEreal Zh_sum = 0.0;
      OEreal Zv_sum = 0.0;
      OEreal Kdp_sum = 0.0;
      OEreal attenh_sum = 0.0;
      OEreal attenv_sum = 0.0;
      OEreal R_sum = 0.0;

      /* Loop over the individual drops in the distribution */
      for (iD = 0; iD < nD; iD++) {
	/* Unnormalized size distribution */
	OEreal N = pow(scattering->diameter.data[iD], mu)
	  * exp( -(3.67+mu) * scattering->diameter.data[iD]
		 /lut->D0.data[iD0]);
	Zh_sum += N * scattering->Zh.data[iT][iD];
	Zv_sum += N * scattering->Zv.data[iT][iD];
	Kdp_sum += N * scattering->Kdp.data[iT][iD];
	attenh_sum += N * scattering->attenh.data[iT][iD];
	attenv_sum += N * scattering->attenv.data[iT][iD];
	R_sum += N * terminal_velocity.data[iD] * volume.data[iD] * 3.6e6;
      }
      /* Calculate the ratios */
      lut->ln_Zh_R.data[iT][iD0] = log(Zh_sum / R_sum);
      lut->ln_Zh_Kdp.data[iT][iD0] = log(Zh_sum / Kdp_sum);
      lut->ln_Zh_attenh.data[iT][iD0] = log(Zh_sum / attenh_sum);
      lut->ln_Zh_attenv.data[iT][iD0] = log(Zh_sum / attenv_sum);
      lut->Zdr.data[iT][iD0] = 10.0 * log10(Zh_sum / Zv_sum);
    }
  }

  /* The linear Zdr of pure hail */
  lut->hail_Zdr_lin = 1.0;
  /* The ratio of Kdp/Ah/Av to linear Zh for pure hail */
  lut->hail_Kdp_Zh = 0.0;/* degree/km / mm6m-3 */
  lut->hail_Ah_Zh = 0.0; /* dB / mm6m-3 */
  lut->hail_Av_Zh = 0.0; /* dB / mm6m-3 */

  oe_free_vector(&terminal_velocity);
  oe_free_vector(&volume);

  return OE_SUCCESS;
}
